"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const ts = require("typescript");
const MODULE_NAME = "inline-class-name";
const FUNCTION_NAME = "className";
function toTemplateExpression(values) {
    if (values.length === 0) {
        return ts.createStringLiteral("");
    }
    let i = 0;
    let s = "";
    while (i < values.length) {
        const v = values[i];
        if (typeof v === "string") {
            if (s !== "") {
                s += " ";
            }
            s += v;
            i++;
        }
        else {
            break;
        }
    }
    if (i === values.length) {
        return ts.createStringLiteral(s);
    }
    s += " ";
    const head = ts.createTemplateHead(s, s);
    const spans = [];
    while (i < values.length) {
        const expr = values[i++];
        s = " ";
        while (i < values.length) {
            const l = values[i];
            if (typeof l === "string") {
                if (s !== " ") {
                    s += " ";
                }
                s += l;
                i++;
            }
            else {
                break;
            }
        }
        if (i < values.length) {
            s += " ";
        }
        else if (s === " ") {
            s = "";
        }
        spans.push(ts.createTemplateSpan(expr, i < values.length ? ts.createTemplateMiddle(s, s) : ts.createTemplateTail(s, s)));
    }
    return ts.createTemplateExpression(head, spans);
}
function stateToBitSet(typeChecker, index, state) {
    let result = 0;
    for (let i = 0; i < state.length; i++) {
        const s = state[i];
        let v;
        if (typeof s === "string") {
            v = index.get(s);
        }
        else {
            const sym = typeChecker.getSymbolAtLocation(s);
            if (sym !== void 0) {
                v = index.get(sym);
            }
        }
        if (v !== void 0) {
            result |= 1 << v;
        }
    }
    return result;
}
function shouldIgnoreState(ignore, state) {
    for (let i = 0; i < ignore.length; i++) {
        const v = ignore[i];
        if ((state & v) === v) {
            return true;
        }
    }
    return false;
}
function createConditional(values, i, prevState, stateIndex, ignoreState, typeChecker) {
    let value;
    let nextState;
    while (i < values.length) {
        value = values[i];
        nextState = prevState.concat([value.name]);
        if (shouldIgnoreState(ignoreState, stateToBitSet(typeChecker, stateIndex, nextState))) {
            i++;
        }
        else {
            break;
        }
    }
    if (i === values.length) {
        return toTemplateExpression(prevState);
    }
    return ts.createParen(ts.createConditional(value.expr, ts.createToken(ts.SyntaxKind.QuestionToken), createConditional(values, i + 1, nextState, stateIndex, ignoreState, typeChecker), ts.createToken(ts.SyntaxKind.ColonToken), createConditional(values, i + 1, prevState, stateIndex, ignoreState, typeChecker)));
}
function inlineClassNameTransformer(program) {
    const typeChecker = program.getTypeChecker();
    return (context) => {
        return (sourceFile) => {
            let inlineClassNameImportModuleStmt;
            sourceFile.statements.forEach((stmt, index) => {
                if (ts.isImportDeclaration(stmt)) {
                    if (ts.isStringLiteral(stmt.moduleSpecifier) && stmt.moduleSpecifier.text === MODULE_NAME) {
                        inlineClassNameImportModuleStmt = stmt;
                    }
                }
            });
            if (inlineClassNameImportModuleStmt === void 0) {
                return sourceFile;
            }
            const inlineClassNames = (node) => {
                if (ts.isCallExpression(node)) {
                    if (node.expression.getText() === FUNCTION_NAME) {
                        const arg1 = node.arguments[0];
                        if (ts.isObjectLiteralExpression(arg1)) {
                            const stateIndex = new Map();
                            const ignoreState = [];
                            const dynamicValues = [];
                            const staticPart = [];
                            arg1.properties.forEach((p) => {
                                if (ts.isPropertyAssignment(p)) {
                                    const name = p.name;
                                    if (ts.isComputedPropertyName(name)) {
                                        const sym = typeChecker.getSymbolAtLocation(name.expression);
                                        if (sym !== void 0) {
                                            stateIndex.set(sym, stateIndex.size);
                                        }
                                    }
                                    else if (ts.isIdentifier(name) || ts.isStringLiteral(name)) {
                                        stateIndex.set(name.text, stateIndex.size);
                                    }
                                    else {
                                        // invalid
                                    }
                                    if (p.initializer.kind === ts.SyntaxKind.TrueKeyword) {
                                        if (ts.isComputedPropertyName(name)) {
                                            staticPart.push(name.expression);
                                        }
                                        else if (ts.isIdentifier(name) || ts.isStringLiteral(name)) {
                                            staticPart.push(name.text);
                                        }
                                    }
                                    else {
                                        dynamicValues.push({
                                            expr: p.initializer,
                                            name: ts.isComputedPropertyName(name) ? name.expression : name.text,
                                        });
                                    }
                                }
                            });
                            if (node.arguments.length > 1) {
                                const arg2 = node.arguments[1];
                                if (ts.isArrayLiteralExpression(arg2)) {
                                    arg2.elements.forEach((e) => {
                                        if (ts.isArrayLiteralExpression(e)) {
                                            let state = 0;
                                            e.elements.forEach((s) => {
                                                if (ts.isStringLiteral(s)) {
                                                    const i = stateIndex.get(s.text);
                                                    if (i !== void 0) {
                                                        state |= 1 << i;
                                                    }
                                                }
                                                else if (ts.isIdentifier(s)) {
                                                    const sym = typeChecker.getSymbolAtLocation(s);
                                                    if (sym !== void 0) {
                                                        const i = stateIndex.get(sym);
                                                        if (i !== void 0) {
                                                            state |= 1 << i;
                                                        }
                                                    }
                                                }
                                            });
                                            if (state !== 0) {
                                                ignoreState.push(state);
                                            }
                                        }
                                    });
                                }
                            }
                            return createConditional(dynamicValues, 0, staticPart, stateIndex, ignoreState, typeChecker);
                        }
                        return node;
                    }
                }
                return ts.visitEachChild(node, inlineClassNames, context);
            };
            const next = ts.visitNode(sourceFile, (node) => ts.visitEachChild(node, inlineClassNames, context));
            return ts.updateSourceFileNode(next, next.statements.filter((v) => v !== inlineClassNameImportModuleStmt));
        };
    };
}
exports.default = inlineClassNameTransformer;
//# sourceMappingURL=index.js.map